(* $Id: termdef.ml,v 1.8 2006/05/27 17:12:47 deraugla Exp $ *)

open C_term;
open Rtdecl;
open Util;
open Xlib;
open Keysym;

type term_event =
  [ TermEvButtonPress of int and int and int and modified int
  | TermEvButtonRelease of int and int and int and modified int
  | TermEvButtonMotion of int and int and int
  | TermEvKeyPress of modified keysym
  | TermEvAnswer of string
  | TermEvExtendHistory
  | TermEvSizeChange of int and int ]
;

type term_event_handler = widget -> term_event -> unit;

value exceptq e =
  exceptq_e where exceptq_e =
    fun
    [ [] -> []
    | [elem :: l] -> if e == elem then l else [elem :: exceptq_e l] ]
;

value union l1 l2 =
  union_rec l1 where union_rec =
    fun
    [ [] -> l2
    | [a :: l] -> if List.mem a l2 then union_rec l else [a :: union_rec l] ]
;

type frozen 'a = [ Ffun of unit -> 'a | Fval of 'a ];
type mlazy 'a = ref (frozen 'a);

type input_state =
  [ IS_normal
  | IS_escape
  | IS_ext of int and list int
  | IS_csi of int and list int
  | IS_csi_question of int and list int
  | IS_escape_sharp
  | IS_escape_lparen
  | IS_escape_rparen ]
;

type term_global_info =
  { dfont : font;
    tgc : gC;
    c_backg : mutable int;
    c_foreg : mutable int;
    c_font : mutable xfont }
and term_local_info =
  { term_gi : term_global_info;
    att_val : attribute_values;
    callb : term_event_handler;
    tfs : array (mlazy font);
    twidth : int;
    theight : int;
    tascent : int;
    tdescent : int;
    max_history_size : int;
    lines : mutable array line;
    nhrow : mutable int;
    nrow : mutable int;
    ncol : mutable int;
    shift : mutable int;
    sreg1 : mutable int;
    sreg2 : mutable int;
    crow : mutable int;
    ccol : mutable int;
    vmask : mutable char;
    foregm : mutable int;
    backgm : mutable int;
    vcrow : mutable int;
    vccol : mutable int;
    vvmask : mutable char;
    vforegm : mutable int;
    vbackgm : mutable int;
    erow1 : mutable int;
    ecol1 : mutable int;
    erow2 : mutable int;
    ecol2 : mutable int;
    scroll_cnt : mutable int;
    scroll_rect : mutable list (int * int * int * int);
    state : mutable input_state;
    tabs : mutable string;
    flags : mutable int }
and line =
  { str : string; vid : string; foreg : array int; backg : array int }
and copy_area =
  { src_row : int;
    src_col : int;
    dst_row : int;
    dst_col : int;
    cp_nrow : int;
    cp_ncol : int }
;

value xor a b = if a then not b else b;
value cland c i = Char.code c land i;
value clor c i = Char.code c lor i;
value clxor c i = Char.code c lxor i;

value f_bld = 1 lsl 0
and f_ita = 1 lsl 1
and f_rev = 1 lsl 2
and f_blk = 1 lsl 3
and f_und = 1 lsl 4
and f_foreg = 1 lsl 5
and f_backg = 1 lsl 6;

value f_gc = f_bld lor f_ita lor f_rev;

value cnt = ref 0;
value new_bit () = (
  let v = cnt.val in
  incr cnt;
  1 lsl v
);

value flg_auto_wrap = new_bit ()
and flg_newline_mode = new_bit ()
and flg_insert_mode = new_bit ()
and flg_appl_num_kp = new_bit ()
and flg_appl_curs_key = new_bit ()
and flg_orig_mode = new_bit ()
and flg_cursor_off = new_bit ()
and flg_reverse_video = new_bit ()
and flg_blink = new_bit ()
and flg_blink_string = new_bit ()
and flg_timeout_set = new_bit ()
and flg_bars = new_bit ()
and flg_trace = new_bit ();

value flg_set li flg = li.flags land flg != 0
and flg_reset li flg = li.flags land flg == 0
and set_flg li flg = li.flags := li.flags lor flg
and reset_flg li flg = li.flags := li.flags land lnot flg
and invert_flg li flg = li.flags := li.flags lxor flg;

value freeze f = ref (Ffun f);
value rec unfreeze v =
  match v.val with
  [ Ffun f -> (v.val := Fval (f ()); unfreeze v)
  | Fval x -> x ]
;

value (term_global_info, get_term_global_info) =
  dynamo_global_info "term_global_info"
    (ref None : ref (option term_global_info))
and (term_local_info, get_term_local_info) =
  dynamo_local_info "term_local_info"
    (ref None : ref (option term_local_info))
;

value term_font =
  [| "*-courier-medium-r-*-14-*"; "*-courier-bold-r-*-14-*";
     "*-courier-medium-o-*-14-*"; "*-courier-bold-o-*-14-*" |]
and term_inter = ref 0
and term_band = ref 2
and term_blink = ref (500, 100);

value set_backg_foreg (xd, li, rev, vid, foreg, backg, font) = (
  let w_foreg =
    if cland vid f_foreg != 0 then foreg
    else
      match li.att_val.foreg_att with
      [ Some (ColorPn c) -> c.pixel
      | _ -> xd.black ]
  in
  let w_backg =
    if cland vid f_backg != 0 then backg
    else
      match li.att_val.backg_att with
      [ Some (ColorPn c) -> c.pixel
      | _ -> xd.white ]
  in
  let (w_foreg, w_backg) =
    if rev then (w_backg, w_foreg) else (w_foreg, w_backg)
  in
  let w_font = font.fid in
  let gi = li.term_gi in
  if gi.c_foreg != w_foreg then (
    xSetForeground (xd.dpy, gi.tgc, w_foreg);
    gi.c_foreg := w_foreg
  )
  else ();
  if gi.c_backg != w_backg then (
    xSetBackground (xd.dpy, gi.tgc, w_backg);
    gi.c_backg := w_backg
  )
  else ();
  if gi.c_font != w_font then (
    xSetFont (xd.dpy, gi.tgc, w_font);
    gi.c_font := w_font
  )
  else ()
);

value expose_row wid li cursor optim_spaces only_blink row bcol ecol =
  let xd = wid.wid_xd in
  let gi = li.term_gi in
  let line = li.lines.(li.nhrow - li.shift + row) in
  expose_loop bcol where expose_loop col1 =
    if col1 < ecol then (
      let vid = String.get line.vid col1 in
      let foreg = Array.get line.foreg col1 in
      let backg = Array.get line.backg col1 in
      let col2 =
        find_end (col1 + 1) where find_end col =
          if col == ecol || String.get line.vid col != vid ||
             Array.get line.foreg col != foreg ||
             Array.get line.backg col != backg
          then
            col
          else find_end (col + 1)
      in
      if cland vid f_blk != 0 then set_flg li flg_blink_string else ();
      if cland vid f_blk == 0 && only_blink then ()
      else
        let ecol =
          let rec find_last_no_space col =
            if col == col1 || String.get line.str (col - 1) != ' ' then col
            else find_last_no_space (col - 1)
          in
          if not optim_spaces || cland vid (f_rev lor f_backg) != 0 then col2
          else find_last_no_space col2
        in
        let bcol =
          let rec find_first_no_space col =
            if col == col1 || String.get line.str col != ' ' then col
            else find_first_no_space (col + 1)
          in
          if not optim_spaces || cland vid (f_rev lor f_backg) != 0 then col1
          else find_first_no_space col1
        in
        if ecol > bcol then (
          let str = String.sub line.str bcol (ecol - bcol) in
          if flg_set li flg_trace then (
            prerr_string ">";
            prerr_int (Char.code vid);
            prerr_string ">";
            prerr_string (if cursor then "|" else str);
            prerr_endline "<<";
            flush stderr
          )
          else ();
          let tband = opt_val term_band.val li.att_val.band_att in
          let tinter = opt_val term_inter.val li.att_val.inter_att in
          let x = tband + bcol * li.twidth in
          let y = tband + row * li.theight + li.tascent in
          let font = unfreeze li.tfs.(cland vid (f_bld lor f_ita)) in
          let rev =
            xor (flg_set li flg_reverse_video) (cland vid f_rev != 0)
          in
          let params = (xd, li, rev, vid, foreg, backg, font) in
          if cursor then (
            set_backg_foreg params;
            let y = y - li.tascent in
            xDrawLine
              (xd.dpy, wid.win, gi.tgc, x, y, x, y + li.theight - tinter - 1)
          )
          else if cland vid f_blk != 0 && flg_reset li flg_blink then
            xClearArea
              (xd.dpy, wid.win, x, y - li.tascent,
               li.twidth * String.length str, li.theight, 0)
          else (
            set_backg_foreg params;
            xDrawImageString
              (xd.dpy, wid.win, gi.tgc, x, y, str, ecol - bcol);
            if cland vid f_und != 0 then
              let y = y + li.tdescent / 2 in
              xDrawLine
                (xd.dpy, wid.win, gi.tgc, x, y,
                 x + li.twidth * String.length str, y)
            else ()
          )
        )
        else ();
      expose_loop col2
    )
    else ()
;

value term_expose_row wid li row bcol ecol =
  expose_row wid li False False False row bcol ecol
;

value term_expose_cursor wid li =
  if li.crow + li.shift < li.nrow && li.ccol < li.ncol then
    expose_row wid li (flg_set li flg_blink && flg_reset li flg_cursor_off)
      False False (li.crow + li.shift) li.ccol (li.ccol + 1)
  else ()
;

value term_expose_blink wid li = (
  term_expose_cursor wid li;
  if flg_set li flg_blink_string then (
    reset_flg li flg_blink_string;
    for row = 0 to li.nrow - 1 do
      expose_row wid li False True True row 0 li.ncol;
    done
  )
  else ()
);


value row_col_of_xy li x y =
  let tband = opt_val term_band.val li.att_val.band_att in
  let row = min li.nrow (max 0 ((y - tband) / li.theight)) in
  let col = min li.ncol (max 0 ((x - tband) / li.twidth)) in
  (row, col)
;

value term_expose wid x y width height = (
  let li = get_term_local_info wid.info in
  let (brow, bcol) = row_col_of_xy li x y in
  let (erow, ecol) =
    row_col_of_xy li (x + width + li.twidth - 1)
      (y + height + li.theight - 1)
  in
  if flg_set li flg_bars then (
    let xd = wid.wid_xd in
    let tb = opt_val term_band.val li.att_val.band_att in
    let bx = tb + bcol * li.twidth in
    let by = tb + brow * li.theight in
    let ex = tb + ecol * li.twidth in
    let ey = tb + erow * li.theight in
    let rec draw_vertical_lines x =
      if x < ex then (
        xDrawLine (xd.dpy, wid.win, xd.gc, x, by, x, ey);
        draw_vertical_lines (x + li.twidth)
      )
      else ()
    and draw_horizontal_lines y = (
      let tinter = opt_val term_inter.val li.att_val.inter_att in
      xDrawLine (xd.dpy, wid.win, xd.gc, bx, y, ex, y);
      let y = y + li.theight - tinter in
      if y < ey then (
        xDrawLine (xd.dpy, wid.win, xd.gc, bx, y, ex, y);
        draw_horizontal_lines (y + tinter)
      )
      else ()
    )
    in
    draw_vertical_lines bx;
    draw_horizontal_lines by
  )
  else ();
  for row = brow to erow - 1 do
    expose_row wid li False True False row bcol ecol;
  done;
  if li.crow >= brow && li.crow < erow && li.ccol >= bcol && li.ccol < ecol
  then
    term_expose_cursor wid li
  else ()
);

value term_soft_reset li = (
  li.state := IS_normal;
  li.vmask := '\000';
  li.foregm := 0;
  li.backgm := 0;
  li.crow := 0;
  li.ccol := 0;
  li.sreg1 := 0;
  li.sreg2 := li.nrow;
  li.erow1 := 0;
  li.ecol1 := 0;
  li.erow2 := 0;
  li.ecol2 := 0;
  li.scroll_cnt := 0;
  li.scroll_rect := [];
  li.flags :=
    flg_auto_wrap lor
    (li.flags land
     (flg_blink lor flg_timeout_set lor flg_bars lor flg_trace));
  for row = li.nhrow to Array.length li.lines - 1 do
    String.fill li.lines.(row).str 0 li.ncol ' ';
    String.fill li.lines.(row).vid 0 li.ncol '\000';
    Array.fill li.lines.(row).backg 0 li.ncol 0;
    Array.fill li.lines.(row).foreg 0 li.ncol 0;
  done;
  for i = 0 to li.ncol - 1 do
    String.set li.tabs i (if i mod 8 == 0 then 't' else ' ');
  done
);

value term_scroll_down wid li nb row nrow = (
  let row = max 0 (min li.nrow row) in
  let nrow = min nrow (li.nrow - row) in
  let nb = min nb nrow in
  let xd = wid.wid_xd in
  let tband = opt_val term_band.val li.att_val.band_att in
  if row + nb < li.nrow then (
    xCopyArea
      (xd.dpy, wid.win, wid.win, xd.gc, tband, tband + row * li.theight,
       li.ncol * li.twidth, (nrow - nb) * li.theight, tband,
       tband + (row + nb) * li.theight);
    li.scroll_cnt := li.scroll_cnt + 1
  )
  else ();
  if row < li.nrow then
    xClearArea
      (xd.dpy, wid.win, tband, tband + row * li.theight, li.ncol * li.twidth,
       nb * li.theight, 0)
  else ()
);

value term_scroll_up wid li nb row nrow = (
  let row = max 0 (min li.nrow row) in
  let nrow = min nrow (li.nrow - row) in
  let nb = min nb nrow in
  let xd = wid.wid_xd in
  let tband = opt_val term_band.val li.att_val.band_att in
  if row + nb < li.nrow then (
    xCopyArea
      (xd.dpy, wid.win, wid.win, xd.gc, tband,
       tband + (row + nb) * li.theight, li.ncol * li.twidth,
       (nrow - nb) * li.theight, tband, tband + row * li.theight);
    li.scroll_cnt := li.scroll_cnt + 1
  )
  else ();
  if row + nrow - nb < li.nrow then
    xClearArea
      (xd.dpy, wid.win, tband, tband + (row + nrow - nb) * li.theight,
       li.ncol * li.twidth, nb * li.theight, 0)
  else ()
);

value show_cursor wid li =
  let xd = wid.wid_xd in
  let rec blinking () = (
    reset_flg li flg_timeout_set;
    if not wid.deleted && flg_reset li flg_cursor_off then (
      invert_flg li flg_blink;
      term_expose_blink wid li;
      set_flg li flg_timeout_set;
      xd.sys_timeout :=
        [(wid,
          if flg_set li flg_blink then fst term_blink.val
          else snd term_blink.val,
          blinking) :: xd.sys_timeout]
    )
    else ()
  )
  in
  if flg_reset li flg_cursor_off then (
    xd.cancel_timeout := exceptq wid xd.cancel_timeout;
    set_flg li flg_blink;
    term_expose_blink wid li;
    if flg_reset li flg_timeout_set then (
      set_flg li flg_timeout_set;
      xd.sys_timeout :=
        [(wid, fst term_blink.val, blinking) :: xd.sys_timeout]
    )
    else ()
  )
  else ()
;

value hide_cursor wid li = (
  let xd = wid.wid_xd in
  reset_flg li flg_blink;
  xd.cancel_timeout := union [wid] xd.cancel_timeout;
  reset_flg li flg_timeout_set;
  term_expose_blink wid li
);

(* Emphasized text *)

value term_invert_zone wid li expose row col1 col2 = (
  let row = max 0 (min (li.nrow - 1) row) in
  let col1 = max 0 (min li.ncol col1)
  and col2 = max 0 (min li.ncol col2) in
  let line = li.lines.(row) in
  for i = col1 to col2 - 1 do
    String.set line.vid i (Char.chr (clxor (String.get line.vid i) f_rev));
  done;
  let row = row - li.nhrow + li.shift in
  if row >= 0 && row < li.nrow && expose then
    term_expose_row wid li row col1 col2
  else ()
);

value display_emphasized_zone wid li expose row1 col1 =
  if row1 == li.erow2 then
    term_invert_zone wid li expose row1 (min col1 li.ecol2)
      (max col1 li.ecol2)
  else (
    let (row1, col1) =
      if row1 < li.erow2 then (row1, col1) else (li.erow2, li.ecol2)
    and (row2, col2) =
      if row1 > li.erow2 then (row1, col1) else (li.erow2, li.ecol2)
    in
    term_invert_zone wid li expose row1 col1 li.ncol;
    for row = row1 + 1 to row2 - 1 do
      term_invert_zone wid li expose row 0 li.ncol;
    done;
    term_invert_zone wid li expose row2 0 col2
  )
;

value adjust_emph_row_col li row col =
  if col == 0 || row == li.nhrow + li.nrow then (row, 0)
  else
    let str = li.lines.(row).str in
    find (col - 1) where find i =
      if i == li.ncol then (row + 1, 0)
      else if String.get str i == ' ' then find (i + 1)
      else (row, col)
;

value term_emphasize_from wid row col = (
  let li = get_term_local_info wid.info in
  let row = li.nhrow - li.shift + max 0 (min (li.nrow - 1) row) in
  let col = max 0 (min col (li.ncol - 1)) in
  let (row, col) = adjust_emph_row_col li row col in
  display_emphasized_zone wid li True li.erow1 li.ecol1;
  li.erow1 := row;
  li.ecol1 := col;
  li.erow2 := row;
  li.ecol2 := col
);

value term_emphasize_to wid row col = (
  let li = get_term_local_info wid.info in
  let row = li.nhrow - li.shift + max 0 (min (li.nrow - 1) row) in
  let col = max 0 (min col (li.ncol - 1)) in
  let (row, col) = adjust_emph_row_col li row col in
  display_emphasized_zone wid li True row col;
  li.erow2 := row;
  li.ecol2 := col
);

value term_get_emphasized wid =
  let li = get_term_local_info wid.info in
  let (row1, col1, row2, col2) =
    if li.erow1 < li.erow2 then (li.erow1, li.ecol1, li.erow2, li.ecol2)
    else if li.erow1 > li.erow2 then (li.erow2, li.ecol2, li.erow1, li.ecol1)
    else if li.ecol1 < li.ecol2 then (li.erow1, li.ecol1, li.erow2, li.ecol2)
    else (li.erow2, li.ecol2, li.erow1, li.ecol1)
  in
  let rec sel_str s row col =
    (**)
    if row == row2 && col == col2 then s
    else
      let str = li.lines.(row).str in
      let last =
        find_last (if row < row2 then li.ncol else col2) where find_last i =
          if i == col then i
          else if String.get str (i - 1) != ' ' then i
          else find_last (i - 1)
      in
      let s = s ^ String.sub str col (last - col) in
      if row == row2 then s else sel_str (s ^ "\n") (row + 1) 0
  in
  (*
      if row == row2 && col == col2 then s
      else
        let str = li.lines.(row).str in
        let last =
          let rec find_last i =
            if i == col then i
            else if String.get str (i - 1) != ' ' then i
            else find_last (i - 1)
          in
          find_last (if row < row2 then li.ncol else col2)
        in
        let s = s ^ String.sub str col (last - col) in
        if row == row2 then s else sel_str (s ^ "\n") (row + 1) 0
  *)
  sel_str "" row1 col1
;

value term_resize wid li nrow ncol = (
  let erow1 = li.erow1
  and ecol1 = li.ecol1 in
  let erow2 = li.erow2
  and ecol2 = li.ecol2 in
  let nrow = max nrow li.nrow in
  let ncol = max ncol li.ncol in
  display_emphasized_zone wid li False li.erow1 li.ecol1;
  let lines =
    make_array (li.nhrow + nrow)
      (fun _ ->
         {str = String.make ncol ' '; vid = String.make ncol '\000';
          foreg = Array.create ncol 0; backg = Array.create ncol 0})
  in
  let tabs = String.make ncol ' ' in
  let min_ncol = min ncol li.ncol in
  for row = 0 to li.nhrow + min nrow li.nrow - 1 do
    String.blit li.lines.(row).str 0 lines.(row).str 0 min_ncol;
    String.blit li.lines.(row).vid 0 lines.(row).vid 0 min_ncol;
    Array.blit li.lines.(row).backg 0 lines.(row).backg 0 min_ncol;
    Array.blit li.lines.(row).foreg 0 lines.(row).foreg 0 min_ncol;
  done;
  for i = 0 to ncol - 1 do
    String.set tabs i (if i mod 8 == 0 then 't' else ' ');
  done;
  String.blit li.tabs 0 tabs 0 min_ncol;
  li.lines := lines;
  li.nrow := nrow;
  li.ncol := ncol;
  li.sreg1 := 0;
  li.sreg2 := nrow;
  li.tabs := tabs;
  li.crow := min li.crow (max 0 (nrow - 1));
  li.ccol := min li.ccol (max 0 (ncol - 1));
  let (row, col) =
    adjust_emph_row_col li (min li.nrow erow1) (min li.ncol ecol1)
  in
  li.erow1 := row;
  li.ecol1 := col;
  li.erow2 := row;
  li.ecol2 := col;
  let (row, col) =
    adjust_emph_row_col li (min li.nrow erow2) (min li.ncol ecol2)
  in
  display_emphasized_zone wid li True row col;
  li.erow2 := row;
  li.ecol2 := col
);
